import java.io.*;
import java.net.*;
import java.util.*;

class iReader extends BufferedReader {
    private BufferedReader in;
    private PrintWriter echo;
    private Stack stk;

    public iReader(BufferedReader IN, PrintWriter ECHO) {
        super(IN);
        in = IN;
        echo = ECHO;
        stk = new Stack();
    }

    public iReader(BufferedReader IN) {
        this(IN, null);
    }

    public String readLine() throws IOException {
        String s;
        try {
            s = (String)stk.pop();
        } catch (EmptyStackException e) {
            s = in.readLine();
        }
        if (echo != null)
            echo.println(s);
        return s;
    }

    public void unreadLine(String S) {
        stk.push(S);
    }

    public boolean stkempty() {
        return stk.empty();
    }
}

class Package implements Comparable {
    public static Hashtable all = new Hashtable();

    public int pid;                     /* package id */
    public int c, r;                    /* package is dropped here, or -1 */
    public int dc, dr;                  /* destination or -1 */
    public int w;                       /* package weight, or 0 */
    public int rid;                     /* robot package is with, or -1 */

    public int dist;                    /* distance for comparison */

    public Package(int PID, int C, int R, int DC, int DR, int W, int RID) {
        pid = PID;
        c = C;
        r = R;
        dc = DC;
        dr = DR;
        w = W;
        rid = RID;
        all.put(new Integer(pid), this);
    }

    public Package(int PID, int C, int R, int DC, int DR, int W) {
        this(PID, C, R, DC, DR, W, -1);
    }

    public Package(int PID, int RID) {
        this(PID, -1, -1, -1, -1, 0, RID);
    }

    public static Package get(int PID) {
        Package p = (Package)all.get(new Integer(PID));
        if (p == null)
            p = new Package(PID, -1);
        return p;
    }

    public void setDetails(int DC, int DR, int W) {
        dc = DC;
        dr = DR;
        w = W;
    }

    public String toString() {
        return "P" + pid + "(" + getc() + "," + getr() + ";" + dc + "," + dr +
            ";" + w + ")" + (rid == -1 ? "" : "(#" + rid + ")") +
            "[" + dist + "]";
    }

    public void disappear() {
        all.remove(new Integer(pid));
    }

    public void moveTo(int C, int R) {
        c = C;
        r = R;
        rid = -1;
        if (c == dc && r == dr)
            disappear();
    }

    public void pickUp(int RID) {
        c = r = -1;
        rid = RID;
    }

    public int getc() {
        return c == -1 ? rid == -1 ? -1 : Robot.get(rid).c : c;
    }

    public int getr() {
        return r == -1 ? rid == -1 ? -1 : Robot.get(rid).r : r;
    }

    public static Vector allAt(int C, int R, boolean carried) {
        Vector pp = new Vector();
        for (Enumeration e = all.elements(); e.hasMoreElements(); ) {
            Package p = (Package)e.nextElement();
            if (carried ?
                (p.getc() == C && p.getr() == R) :
                (p.c == C && p.r == R))
                pp.add(p);
        }
        return pp;
    }

    public int compareTo(Object o) {
        Package p = (Package)o;
        if (dist != p.dist)
            return dist - p.dist;
        else
            return w - p.w;
    }
}

class Robot {
    public static Hashtable all = new Hashtable();

    public int rid;                     /* robot id */
    public int c, r;                    /* location */
    public Hashtable carrying;          /* packages being carried */

    public Robot(int RID, int C, int R) {
        rid = RID;
        c = C;
        r = R;
        carrying = new Hashtable();
        all.put(new Integer(rid), this);
    }

    public static Robot get(int RID) {
        Robot b = (Robot)all.get(new Integer(RID));
        if (b == null)
            b = new Robot(RID, 1, 1);
        return b;
    }

    public String toString() {
        return "#" + rid + "(" + c + "," + r + ";" + carrying + ")";
    }

    public void moveTo(int C, int R) {
        c = C;
        r = R;
    }

    public void moveBy(int DC, int DR) {
        moveTo(c + DC, r + DR);
    }

    public void move(char C) {
        switch (C) {
            case 'n': case 'N': moveBy( 0, +1); break;
            case 's': case 'S': moveBy( 0, -1); break;
            case 'e': case 'E': moveBy(+1,  0); break;
            case 'w': case 'W': moveBy(-1,  0); break;
        }
    }

    public void pickUp(int PID) {
        Package p = Package.get(PID);
        carrying.put(new Integer(PID), p);
        p.pickUp(rid);
    }

    public void drop(int PID) {
        Package p = Package.get(PID);
        carrying.remove(new Integer(PID));
        p.moveTo(c, r);
    }

    public void disappear() {
        all.remove(new Integer(rid));
    }
}

class idiot {
    static Socket sock;
    static iReader sin;
    static PrintWriter sout;
    static int cols, rows;
    static char[][] map;
    static int[][] mappkgs;
    static int id, icap, cap, money;
    static Robot us;
    static int lastc, lastr;
    static boolean kenaPushed;

    public static void main(String[] ARGS) throws Exception {
        init(ARGS[0], Integer.parseInt(ARGS[1]));
        while (true) {
            readPackages();
            think();
            lastc = us.c;
            lastr = us.r;
            kenaPushed = false;
            readResponse();
            if ((us.c != lastc + (
                    lastCommand.equalsIgnoreCase("move e") ? 1 :
                    lastCommand.equalsIgnoreCase("move w") ? -1 : 0)) ||
                (us.r != lastr + (
                    lastCommand.equalsIgnoreCase("move n") ? 1 :
                    lastCommand.equalsIgnoreCase("move s") ? -1 : 0)))
                kenaPushed = true;
            if (kenaPushed)
                defbid += 2;
            else if (defbid > 1)
                defbid--;
        }
    }

    static void init(String HOST, int PORT) throws Exception {
        String temp;
        StringTokenizer st;
        int c, r;

        sock = new Socket(InetAddress.getByName(HOST), PORT);
        sin = new iReader(new BufferedReader(new InputStreamReader(
            sock.getInputStream()), 1)/*, new PrintWriter(System.out, true)*/);
        sout = new PrintWriter(new OutputStreamWriter(sock.getOutputStream()),
            true);

        sout.println("Player");

        st = new StringTokenizer(sin.readLine());
        cols = Integer.parseInt(st.nextToken());
        rows = Integer.parseInt(st.nextToken());
        map = new char[cols + 2][rows + 2];
        mappkgs = new int[cols + 2][rows + 2];
        for (r = rows + 1; r >= 0; r--)
            map[cols + 1][r] = map[0][r] = '#';
        for (c = cols; c >= 0; c--)
            map[c][rows + 1] = map[c][0] = '#';
        for (r = 1; r <= rows; r++) {
            temp = sin.readLine();
            for (c = 1; c <= cols; c++)
                if ((map[c][r] = temp.charAt(c - 1)) == '@')
                    mappkgs[c][r] = Integer.MAX_VALUE / 2;
        }

        st = new StringTokenizer(sin.readLine());
        id = Integer.parseInt(st.nextToken());
        icap = cap = Integer.parseInt(st.nextToken());
        money = Integer.parseInt(st.nextToken());

        st = new StringTokenizer(sin.readLine(), "# XY");
        while (st.hasMoreTokens())
            Robot.get(Integer.parseInt(st.nextToken())).moveTo(
                Integer.parseInt(st.nextToken()),
                Integer.parseInt(st.nextToken()));

        us = Robot.get(id);
    }

    static void readPackages() throws Exception {
        Package p;

        StringTokenizer st = new StringTokenizer(sin.readLine());
        mappkgs[us.c][us.r] = st.countTokens() / 4;
        try {
            while (st.hasMoreTokens()) {
                p = Package.get(Integer.parseInt(st.nextToken()));
                p.setDetails(
                        Integer.parseInt(st.nextToken()),
                        Integer.parseInt(st.nextToken()),
                        Integer.parseInt(st.nextToken()));
                p.moveTo(us.c, us.r);
            }
        } catch (Exception e) {
            while (sin.readLine() != null)
                ;
            throw e;
        }
    }

    static void deduct(int BID) {
        money -= Math.abs(BID);
    }

    static String lastCommand = "";

    static void doCommand(int BID, String CMD, String S) {
        if (gotThreat(CMD + " " + S))
            BID++;
        sout.println(BID + " " + (lastCommand = CMD + " " + S));
        deduct(BID);
    }

    static iReader kbd = new iReader(new BufferedReader(
        new InputStreamReader(System.in), 1)/*,
        new PrintWriter(System.out, true)*/); /* DEBUG */

    static void think() throws Exception {
        killDeadBots();
        if (map[us.c][us.r] == '@')
            map[us.c][us.r] = '.';
        findNextTarget();
        /*System.out.println("[" + drow + "," + dcol + "]");*/
        manualThink();
    }

    static boolean within(int lb, int x, int ub) {
        return lb <= x && x <= ub;
    }

    static boolean inside(int c, int r) {
        return within(1, c, cols) && within(1, r, rows);
    }

    static boolean visitable(int c, int r) {
        return map[c][r] == '.' || map[c][r] == '@';
    }

    static boolean okay(int c, int r) {
        return inside(c, r) && visitable(c, r);
    }

    static void killDeadBots() {
        for (Enumeration e = Robot.all.elements(); e.hasMoreElements(); ) {
            Robot b = (Robot)e.nextElement();
            if (!okay(b.c, b.r))
                b.disappear();
        }
    }

    static int defbid = 1, dcol = -1, drow = -1;

    static void paint(int[][] mapz, int cc, int cr, int cv) {
        if (mapz[cc][cr] > cv) {
            mapz[cc][cr] = cv;
            paint(mapz, cc, cr + 1, cv + 1);
            paint(mapz, cc, cr - 1, cv + 1);
            paint(mapz, cc + 1, cr, cv + 1);
            paint(mapz, cc - 1, cr, cv + 1);
        }
    }

    static int statevar = 0;
    static int[][] mapd = null;
    static int pdcol = -1, pdrow = -1;

    static int distance(int fc, int fr, int tc, int tr) {
        if (!(okay(fc, fr) && okay(tc, tr)))
            return Integer.MAX_VALUE;

        int c, r;
        String s;

        if (tc != pdcol || tr != pdrow) {
            mapd = new int[cols + 2][rows + 2];

            for (c = cols + 1; c >= 0; c--)
                for (r = rows + 1; r >= 0; r--)
                    mapd[c][r] = visitable(c, r) ? Integer.MAX_VALUE : -1;

            paint(mapd, tc, tr, 0);

            for (c = cols + 1; c >= 0; c--)
                for (r = rows + 1; r >= 0; r--)
                    if (mapd[c][r] < 0)
                        mapd[c][r] = Integer.MAX_VALUE;
        }

        return mapd[fc][fr];
    }

    static void calcDist(int c, int r) {
        distance(c, r, c, r);
    }

    static String dijkstra() {
        String s;
        int c;

        if (distance(us.c, us.r, dcol, drow) >= Integer.MAX_VALUE)
            return "";

        s = "";
        c = Integer.MAX_VALUE;
        if (mapd[us.c][us.r + 1] < c) {
            c = mapd[us.c][us.r + 1];
            s = "n";
        } else if (mapd[us.c][us.r + 1] == c)
            s += "n";
        if (mapd[us.c][us.r - 1] < c) {
            c = mapd[us.c][us.r - 1];
            s = "s";
        } else if (mapd[us.c][us.r - 1] == c)
            s += "s";
        if (mapd[us.c + 1][us.r] < c) {
            c = mapd[us.c + 1][us.r];
            s = "e";
        } else if (mapd[us.c + 1][us.r] == c)
            s += "e";
        if (mapd[us.c - 1][us.r] < c) {
            c = mapd[us.c - 1][us.r];
            s = "w";
        } else if (mapd[us.c - 1][us.r] == c)
            s += "w";
        if (mapd[us.c][us.r] < c)
            s = "";

        return s;
    }

    static void findNextTarget() {
        int c, r, min;

        /*if (inside(drow, dcol))
            return;*/ /* got target */

        if (statevar == 0 && us.carrying.isEmpty()) {  /* empty-handed */
            calcDist(us.c, us.r);

            min = Integer.MAX_VALUE;
            for (c = cols; c > 0; c--)
                for (r = rows; r > 0; r--)
                    if (mappkgs[c][r] > 0 && mapd[c][r] < min) {
                        min = mapd[c][r];
                        dcol = c;
                        drow = r;
                    }
        } else if (statevar == 0) {     /* something at hand */
            Object[] opkgs = us.carrying.values().toArray();
            calcDist(us.c, us.r);
            for (int i = 0; i < opkgs.length; i++) {
                Package pkg = (Package)opkgs[i];
                pkg.dist = mapd[pkg.dc][pkg.dr];
            }
            Arrays.sort(opkgs);
            Package pkg = (Package)opkgs[0];
            dcol = pkg.dc;
            drow = pkg.dr;
        }
    }

    static boolean gotThreat(int c, int r, int dc, int dr) {
        Robot b;
        for (Enumeration e = Robot.all.elements(); e.hasMoreElements(); ) {
            b = (Robot)e.nextElement();
            if (b.c == c && b.r == r)
                ;
            else if (Math.abs(b.c - c) + Math.abs(b.r - r) == 1)
                return true;
            else if (Math.abs(b.c - dc) + Math.abs(b.r - dr) == 1)
                return true;
        }
        return false;
    }

    static boolean gotThreat(String cmd) {
        return gotThreat(us.c, us.r,
            us.c + (cmd.equalsIgnoreCase("move e") ? 1 :
                    cmd.equalsIgnoreCase("move w") ? -1 : 0),
            us.r + (cmd.equalsIgnoreCase("move n") ? 1 :
                    cmd.equalsIgnoreCase("move s") ? -1 : 0));
    }

    static void manualThink() throws Exception {
        boolean printmap = false;
        while (true) {
            if (printmap) {
                char[][] mapz = new char[cols + 2][rows + 2];
                int c, r;
                Enumeration e;
                Robot b;

                for (c = cols + 1; c >= 0; c--)
                    for (r = rows + 1; r >= 0; r--) {
                        if (mappkgs[c][r] <= 0)
                            mapz[c][r] = map[c][r];
                        else if (mappkgs[c][r] < 10)
                            mapz[c][r] = (char)(mappkgs[c][r] + '0');
                        else if (mappkgs[c][r] < 36)
                            mapz[c][r] = (char)(mappkgs[c][r] - 10 + 'A');
                        else
                            mapz[c][r] = '%';
                    }

                for (e = Robot.all.elements(); e.hasMoreElements(); ) {
                    b = (Robot)e.nextElement();
                    mapz[b.c][b.r] = '*';
                }
                mapz[us.c][us.r] = '+';

                System.out.println(Package.allAt(us.c, us.r, true));
                System.out.println();
                for (r = rows; r > 0; r--) {
                    for (c = 1; c <= cols; c++)
                        System.out.print(mapz[c][r]);
                    System.out.println(" " + r);
                }
                for (c = 1; c <= cols; c++) {
                    for (r = c; r > 0 && r % 10 == 0; r /= 10) ;
                    System.out.print(r % 10);
                }
                System.out.println();
                System.out.print("WxH=" + cols + "x" + rows + " #" + id +
                    " C=" + cap + " $" + money);
                for (e = Robot.all.elements(); e.hasMoreElements(); )
                    System.out.print(" " + e.nextElement());
                System.out.println();

                printmap = false;
            }

            /*System.out.println("Order?");*/

            try {
                int i;
                String dijk = dijkstra();
                if (dijk.length() == 0) {
                    dcol = drow = -1;
                    statevar++;
                } else {
                    char nextmove = dijk.charAt(
                        (int)Math.floor(Math.random() * dijk.length()));
                    kbd.unreadLine(nextmove + "");
                }
                if (true || statevar == 1) {
                    String topickup = "";
                    int rcap = cap;
                    Object[] opkgs =
                        Package.allAt(us.c, us.r, false).toArray();
                    calcDist(us.c, us.r);
                    while (opkgs.length > 0) {
                        for (i = 0; i < opkgs.length; i++) {
                            Package pkg = (Package)opkgs[i];
                            pkg.dist = mapd[pkg.dc][pkg.dr];
                        }
                        Arrays.sort(opkgs);
                        Package pkg = (Package)opkgs[0];
                        if (pkg.w <= rcap) {
                            topickup += " " + pkg.pid;
                            rcap -= pkg.w;
                            if (rcap == 0)
                                break;
                        }
                        calcDist(pkg.c, pkg.r);
                        Vector vpkgs = new Vector(Arrays.asList(opkgs));
                        vpkgs.remove(0);
                        opkgs = vpkgs.toArray();
                    }
                    if (topickup.length() > 0)
                        kbd.unreadLine("Pick" + topickup);
                    statevar = 0;
                }
                String todrop = "";
                for (Enumeration e = us.carrying.elements();
                    e.hasMoreElements(); ) {
                    Package p = (Package)e.nextElement();
                    if (p.dc == us.c && p.dr == us.r)
                        todrop += " " + p.pid;
                }
                if (todrop.length() > 0)
                    kbd.unreadLine("Drop" + todrop);
                if (kbd.stkempty())
                    kbd.unreadLine("1 Pick");
                String s = kbd.readLine();

                if (s == null)
                    s = "1 Pick";
                if (s.length() == 0) {
                    printmap = true;
                    continue;
                }
                if (s.startsWith("=")) {
                    kbd.unreadLine("");
                    for (i = s.length() - 1; i > 0; i--)
                        if ("NEWSnews".indexOf(s.charAt(i)) >= 0)
                            kbd.unreadLine(s.substring(i, i + 1));
                }
                if (s.equalsIgnoreCase("quit") ||
                    s.equalsIgnoreCase("exit"))
                    s = "0 Drop game";

                StringTokenizer st = new StringTokenizer(s);
                String sbid = st.nextToken();
                if (sbid == null)
                    System.out.println(sbid);   /* dummy. seems necessary */
                int bid = defbid;
                try {
                    bid = Integer.parseInt(sbid);
                    sbid = st.nextToken();
                } catch (Exception e) {
                }
                char cmd = sbid.toLowerCase().charAt(0);
                s = "";
                while (st.hasMoreTokens())
                    s += " " + st.nextToken();
                if (s.length() > 0)
                    s = s.substring(1);

                if (cmd == 'm' && s.length() == 1 &&
                    "NEWSnews".indexOf(s.charAt(0)) >= 0) {
                    doCommand(bid, "Move", s.toUpperCase());
                    break;
                } else if ("news".indexOf(cmd) >= 0) {
                    doCommand(bid, "Move", ("" + cmd).toUpperCase());
                    break;
                } else if (cmd == 'p') {
                    doCommand(bid, "Pick", s);
                    break;
                } else if (cmd == 'd') {
                    doCommand(bid, "Drop", s);
                    break;
                } else if (cmd == 'g') {
                    System.out.println("G(" + bid + " " + s + ")");
                    if (defbid != bid) {
                        System.out.println("Default bid set to " + bid);
                        defbid = bid;
                    }
                    try {
                        st = new StringTokenizer(s);
                        dcol = Integer.parseInt(st.nextToken());
                        drow = Integer.parseInt(st.nextToken());
                        System.out.println("Target set at (" + dcol + "," +
                            drow + ")");
                    } catch (Exception e) {
                        System.out.println("No target set");
                        dcol = drow = -1;
                    }
                }
            } catch (Exception e) {
            }
        }
    }

    static void readResponse() throws Exception {
        int rid = id, pid;
        Hashtable mentioned = new Hashtable();

        StringTokenizer st = new StringTokenizer(sin.readLine());
        while (st.hasMoreTokens()) {
            String s = st.nextToken();
            if (s.charAt(0) == '#') {
                rid = Integer.parseInt(s.substring(1));
                mentioned.put(new Integer(rid), Robot.get(rid));
            } else {
                Robot b = Robot.get(rid);
                switch (s.charAt(0)) {
                    case '#':
                        rid = Integer.parseInt(s.substring(1));
                        mentioned.put(new Integer(rid), b);
                        break;
                    case 'N': case 'S': case 'E': case 'W':
                        b.move(s.charAt(0));
                        break;
                    case 'P':
                        pid = Integer.parseInt(st.nextToken());
                        b.pickUp(pid);
                        mappkgs[b.c][b.r]--;
                        if (rid == id)
                            cap -= Package.get(pid).w;
                        break;
                    case 'D':
                        pid = Integer.parseInt(st.nextToken());
                        mappkgs[b.c][b.r]++;
                        if (rid == id) {
                            cap += Package.get(pid).w;
                            if ("dD".indexOf(lastCommand.charAt(0)) < 0)
                                kenaPushed = true;
                        }
                        b.drop(pid);
                        break;
                    case 'X':
                        b.c = Integer.parseInt(st.nextToken());
                        break;
                    case 'Y':
                        b.r = Integer.parseInt(st.nextToken());
                        break;
                }
            }
        }

        for (Enumeration e = Robot.all.elements(); e.hasMoreElements(); ) {
            Robot b = (Robot)e.nextElement();
            if (!mentioned.containsKey(new Integer(b.rid)))
                b.disappear();
        }
    }
}
